@@ -30,7 +30,7 @@
 
 ;;; version.
 
-(defconst iiimcf-server-control-version "0.2 (Takase)")
+(defconst iiimcf-server-control-version "0.3 (Yodo)")
 
 ;;; Customizable options.
 
@@ -57,11 +57,18 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
   :tag "Username for IIIMCF Server Control"
   :group 'iiimcf-server-control :type 'string)
 
-(defcustom iiimcf-server-control-default-language "ja"
-  "*Use this language by default."
+(defcustom iiimcf-server-control-default-language nil
+  "*Use this language by default.
+If nil, use the first candidate sent by a server."
   :tag "Default language for IIIMCF Server Control"
   :group 'iiimcf-server-control :type 'string)
 
+(defcustom iiimcf-server-control-default-input-method nil
+  "*Use this input method by default.
+If nil, don't apply input method preference to a server."
+  :tag "Default input method for IIIMCF Server Control"
+  :group 'iiimcf-server-control :type 'string)
+
 (defcustom iiimcf-server-control-async-invocation-p t
   "*Receives and responds incomming messages asynchronously, if non-nil."
   :tag "Aaync invocation flag for IIIMCF Server Control"
@@ -79,10 +86,47 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
   :tag "Preedit close string for IIIMCF Server Control"
   :group 'iiimcf-server-control :type 'string)
 
+(defface iiimcf-server-control-preedit-reverse-face
+  '((((class color)
+      (background dark))
+     (:foreground "black" :background "DarkSlateGray1"))
+    (((class color)
+      (background light))
+     (:foreground "black" :background "DarkSlateGray3"))
+    (t (:inverse-video t)))
+  "Face used for preedit reverse attribute"
+  :group 'iiimcf-server-control)
+
+(defface iiimcf-server-control-preedit-underline-face
+  '((t (:underline t)))
+  "Face used for preedit underline attribute"
+  :group 'iiimcf-server-control)
+
+(defface iiimcf-server-control-preedit-highlight-face
+  '((((class color)
+      (background dark))
+     (:bold t :foreground "black" :background "SeaGreen1"))
+    (((class color)
+      (background light))
+     (:bold t :foreground "black" :background "SeaGreen3"))
+    (t (:bold t)))
+  "Face used for preedit highlight attribute"
+  :group 'iiimcf-server-control)
+
+(defface iiimcf-server-control-preedit-warning-face
+  '((t (:bold t :foreground "yellow" :background "red")))
+  "Face used for warning for users"
+  :group 'iiimcf-server-control)
+
 (defcustom iiimcf-server-control-preedit-face-alist
-  '((reverse region ("[" . "]"))
-    (underline underline ("_" . "_"))
-    (highlight bold ("+" . "+")))
+  '((reverse iiimcf-server-control-preedit-reverse-face
+	     ("[" . "]"))
+    (underline iiimcf-server-control-preedit-underline-face
+	       ("_" . "_"))
+    (highlight iiimcf-server-control-preedit-highlight-face
+	       ("+" . "+"))
+    (unknown-ucs iiimcf-server-control-preedit-warning-face
+		 ("X" . "X")))
   "*A-list used for decorating preedit text."
   :tag "Preedit decoration settings for IIIMCF Server Control"
   :group 'iiimcf-server-control
@@ -101,31 +145,38 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 
 ;;; internal variables
 
+(defvar iiimcf-server-control-enable nil)
+(make-variable-buffer-local 'iiimcf-server-control-enable)
+(put 'iiimcf-server-control-enable 'permanent-local t)
+(defvar iiimcf-server-control-preedit-mode nil)
+(make-variable-buffer-local 'iiimcf-server-control-preedit-mode)
+(put 'iiimcf-server-control-preedit-mode 'permanent-local t)
+
 (defvar iiimcf-server-control-com-id nil)
 (defvar iiimcf-server-control-im-id nil)
 (defvar iiimcf-server-control-ic-id nil)
 (defvar iiimcf-server-control-dynamic-event-flow-p nil)
-(defvar iiimcf-server-control-input-method-filtered-events nil)
 (defvar iiimcf-server-control-editing-buffer nil
   "The buffer where a user is inputting currently.")
 (defvar iiimcf-server-control-preedit "")
 (defvar iiimcf-server-control-preedit-marker (make-marker))
-(defvar iiimcf-server-control-caret-overlay
-  (make-overlay 0 0))
-(overlay-put iiimcf-server-control-caret-overlay
-	     'face 'modeline)
+;; (defvar iiimcf-server-control-caret-overlay
+;;   (make-overlay 0 0))
+;; (overlay-put iiimcf-server-control-caret-overlay
+;;	     'face 'modeline)
 (defvar iiimcf-server-control-maintaining-preedit-text nil)
 
 (defvar iiimcf-server-control-mode-line "")
 (make-variable-buffer-local 'iiimcf-server-control-mode-line)
-(if (not (memq 'iiimcf-server-control-mode-line mode-line-format))
+(if (null (memq 'iiimcf-server-control-mode-line mode-line-format))
     (let ((ll mode-line-format)
-	  gen elem slot)
+	  gen elem)
       (setq-default
        mode-line-format
        (catch 'tag
 	 (while (setq elem (car ll))
-	   (if (or (eq elem 'mode-line-mule-info)
+	   (if (or (memq elem '(mode-line-mule-info
+				modeline-multibyte-status))
 		   (equal elem "-%-"))
 	       (throw 'tag (nconc (nreverse gen)
 				  (list 'iiimcf-server-control-mode-line)
@@ -137,38 +188,62 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 (defvar iiimcf-server-control-current-lookup-choice nil)
 (defvar iiimcf-server-control-lookup-choice-configuration nil)
 
+;;; private database
+
+(defvar iiimcf-server-control-com-id-plists nil)
+
+(defun iiimcf-server-control-register-com-id (com-id)
+  (if (null (assq com-id iiimcf-server-control-com-id-plists))
+      (setq iiimcf-server-control-com-id-plists
+	    (cons (list com-id)
+		  iiimcf-server-control-com-id-plists))))
+
+(defun iiimcf-server-control-unregister-com-id (com-id)
+  (let ((slot (assq com-id iiimcf-server-control-com-id-plists)))
+    (if slot
+	(setq iiimcf-server-control-com-id-plists
+	      (delq slot
+		    iiimcf-server-control-com-id-plists)))))
+
+(defsubst iiimcf-server-control-com-id-get (com-id prop)
+  (plist-get (cdr (assq com-id iiimcf-server-control-com-id-plists))
+	     prop))
+
+(defsubst iiimcf-server-control-com-id-put (com-id prop val)
+  (let ((slot (assq com-id iiimcf-server-control-com-id-plists)))
+    (if slot
+	(setcdr slot (plist-put (cdr slot) prop val)))))
+
 ;;; keymap
 
-(defvar iiimcf-server-control-minor-mode-keymap
-  (let ((map (make-keymap)))
-    (define-key map "\C-c\C-\\\C-l" 'iiimcf-server-control-change-language)
-    (define-key map "\C-c\C-\\\C-c" 'iiimcf-server-control-switch-ic-attribute)
-    (define-key map "\C-c\C-\\\C-s" 'iiimcf-server-control-change-iiim-server)
-    map))
-
-(defvar iiimcf-server-control-keymap
+;; keymap for initial state.
+(defvar iiimcf-server-control-initial-state-keymap
   (let ((map (make-keymap))
-	(i 0))
+	(i 32))
     (while (< i 127)
       (define-key map (char-to-string i)
 	'iiimcf-server-control-keyforward)
       (setq i (1+ i)))
+    ;; For our own features.
+    (define-key map "\C-c\C-\\\C-l" 'iiimcf-server-control-change-language)
+    (define-key map "\C-c\C-\\\C-i" 'iiimcf-server-control-change-input-method)
+    (define-key map "\C-c\C-\\\C-s" 'iiimcf-server-control-change-iiim-server)
+    map))
+
+;; keymap for preedit state
+(defvar iiimcf-server-control-preedit-state-keymap
+  (let ((map (make-keymap))
+	(i 0))
+    (while (<= i 127)
+      (if (/= i ?\x1B) ; skip ESC
+	  (define-key map (char-to-string i)
+	    'iiimcf-server-control-keyforward))
+      (setq i (1+ i)))
     (setq i 128)
     (while (< i 256)
       (define-key map (vector i)
 	'iiimcf-server-control-keyforward)
       (setq i (1+ i)))
-    (define-key map "\C-m" 'iiimcf-server-control-keyforward)
-    (define-key map "\C-p" 'iiimcf-server-control-keyforward)
-    (define-key map "\C-n" 'iiimcf-server-control-keyforward)
-    (define-key map "\C-b" 'iiimcf-server-control-keyforward)
-    (define-key map "\C-f" 'iiimcf-server-control-keyforward)
-    (define-key map "\C-a" 'iiimcf-server-control-keyforward)
-    (define-key map "\C-e" 'iiimcf-server-control-keyforward)
-    (define-key map "\C-d" 'iiimcf-server-control-keyforward)
-    (define-key map "\C-k" 'iiimcf-server-control-keyforward)
-    (define-key map "\C-h" 'iiimcf-server-control-keyforward)
-    (define-key map "\177" 'iiimcf-server-control-keyforward)
     (define-key map [delete] 'iiimcf-server-control-keyforward)
     (define-key map [backspace] 'iiimcf-server-control-keyforward)
     (define-key map [return] 'iiimcf-server-control-keyforward)
@@ -176,16 +251,56 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
     (define-key map [down] 'iiimcf-server-control-keyforward)
     (define-key map [left] 'iiimcf-server-control-keyforward)
     (define-key map [right] 'iiimcf-server-control-keyforward)
-    (define-key map [S-up] 'iiimcf-server-control-keyforward)
-    (define-key map [S-down] 'iiimcf-server-control-keyforward)
-    (define-key map [S-left] 'iiimcf-server-control-keyforward)
-    (define-key map [S-right] 'iiimcf-server-control-keyforward)
-    (define-key map [C-up] 'iiimcf-server-control-keyforward)
-    (define-key map [C-down] 'iiimcf-server-control-keyforward)
-    (define-key map [C-left] 'iiimcf-server-control-keyforward)
-    (define-key map [C-right] 'iiimcf-server-control-keyforward)
+    (define-key map [(shift up)] 'iiimcf-server-control-keyforward)
+    (define-key map [(shift down)] 'iiimcf-server-control-keyforward)
+    (define-key map [(shift left)] 'iiimcf-server-control-keyforward)
+    (define-key map [(shift right)] 'iiimcf-server-control-keyforward)
+    (define-key map [(control up)] 'iiimcf-server-control-keyforward)
+    (define-key map [(control down)] 'iiimcf-server-control-keyforward)
+    (define-key map [(control left)] 'iiimcf-server-control-keyforward)
+    (define-key map [(control right)] 'iiimcf-server-control-keyforward)
+    (set-keymap-parent map iiimcf-server-control-initial-state-keymap)
     map)
-  "Keymap used for forwarding keyevents to IIIM server side.")
+  "Keymap used for forwarding keyevents to IIIM server side in preedit state.")
+
+;;; IIIMCF event handler.
+
+(defconst iiimcf-server-control-com-level-handler-alist
+  '((iiimcf-server-control-setup-event-flow-mode . iiimp-im-register-trigger-keys)
+    (iiimcf-server-control-maintain-language-list . iiimp-im-connect-reply)
+    (iiimcf-server-control-maintain-input-method-list . iiimp-im-setimvalues)))
+
+(defconst iiimcf-server-control-ic-level-handler-alist
+  '((iiimcf-server-control-trigger-notify . iiimp-im-trigger-notify)
+    (iiimcf-server-control-maintain-preedit-text . iiimp-im-preedit-draw)
+    (iiimcf-server-control-commit-string . iiimp-im-commit-string)
+    (iiimcf-server-control-forward-event . iiimp-im-forward-event)
+    (iiimcf-server-control-forward-event . iiimp-im-forward-event-with-operations)
+    (iiimcf-server-control-prepare-lookup-choice . iiimp-im-lookup-choice-start)
+    (iiimcf-server-control-draw-lookup-choice . iiimp-im-lookup-choice-draw)
+    (iiimcf-server-control-redraw-lookup-choice . iiimp-im-lookup-choice-process)
+    (iiimcf-server-control-clear-lookup-choice . iiimp-im-lookup-choice-done)
+    (iiimcf-server-control-draw-status . iiimp-im-status-draw)
+    (iiimcf-server-control-draw-aux-data . iiimp-im-aux-draw)))
+
+(defun iiimcf-server-control-register-handlers
+  (alist com-id &optional im-id ic-id)
+  (let (elem)
+    (while (setq elem (car alist))
+      (iiimcf-register-handler
+       (car elem)
+       (if (and im-id ic-id)
+	   (list (cdr elem) com-id im-id ic-id)
+	 (list (cdr elem) com-id)))
+      (setq alist (cdr alist)))))
+
+(defun iiimcf-server-control-remove-handlers (alist)
+  (while alist
+    (iiimcf-remove-handler
+     (car (car alist)) nil)
+    (setq alist (cdr alist))))
+
+;;;
 
 (defun iiimcf-server-control-parse-hostname (hostname)
   (if (string-match ":" hostname)
@@ -206,7 +321,7 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 		feedbacks (cdr feedbacks))
 	  (if slot
 	      (setq result (cons (nth 1 slot) result))))
-	result)
+	(nreverse result))
     nil))
 
 (defun iiimcf-server-control-setup-event-flow-mode (mes)
@@ -219,6 +334,85 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
      (iiimcf-server-control-shutdown)
      (signal (car err) (cdr err)))))
 
+(defun iiimcf-server-control-get-ic-attribute (com-id im-id)
+  (let ((lang iiimcf-server-control-default-language)
+	(im iiimcf-server-control-default-input-method))
+    (if (null lang)
+	(setq lang
+	      (car
+	       (iiimcf-server-control-com-id-get
+		com-id 'language-list))))
+    (nconc
+     (list (cons 'input-language lang))
+     ;; Currently, disabled, because the latest version of htt_server
+     ;; accepts not attribid but string of the engin name.
+     ;; This behavior clearly violates the specification.
+     ;; (if im (list (cons 'input-method im)) nil)
+     )))
+
+(defun iiimcf-server-control-ic-setup (com-id im-id)
+  (let (ic-id mes)
+    (iiimp-send-message
+     'iiimp-im-createic
+     com-id im-id
+     (iiimcf-server-control-get-ic-attribute
+      com-id im-id))
+    (setq mes
+	  (iiimcf-message-manager
+	   com-id
+	   (list
+	    (list 'iiimp-im-createic-reply
+		  com-id im-id))))
+    (setq ic-id (nth 3 mes))
+    (iiimp-send-message 'iiimp-im-resetic com-id im-id ic-id)
+    (iiimcf-message-manager
+     com-id
+     (list
+      (list 'iiimp-im-resetic-reply
+	    com-id im-id ic-id)))
+    ic-id))
+
+(defun iiimcf-server-control-ic-destruct (com-id im-id ic-id)
+  (iiimp-send-message 'iiimp-im-resetic com-id im-id ic-id)
+  (iiimcf-message-manager
+   com-id
+   (list
+    (list 'iiimp-im-resetic-reply
+	  com-id im-id ic-id)))
+  (iiimp-send-message 'iiimp-im-destroyic com-id im-id ic-id)
+  (iiimcf-message-manager
+   com-id
+   (list
+    (list 'iiimp-im-destroyic-reply
+	  com-id im-id ic-id))))
+
+(defmacro iiimcf-server-control-ic-operation
+  (x-com-id x-im-id x-ic-id &rest form)
+  `(progn
+     (iiimp-send-message
+      'iiimp-im-seticfocus ,x-com-id ,x-im-id ,x-ic-id)
+     (iiimcf-message-manager
+      ,x-com-id
+      (list
+       (list
+	'iiimp-im-seticfocus-reply
+	,x-com-id ,x-im-id ,x-ic-id)))
+     ,@form
+     (iiimp-send-message
+      'iiimp-im-unseticfocus ,x-com-id ,x-im-id ,x-ic-id)
+     (iiimcf-message-manager
+      ,x-com-id
+      (list
+       (list
+	'iiimp-im-unseticfocus-reply
+	,x-com-id ,x-im-id ,x-ic-id)))))
+
+(defun iiimcf-server-control-dispatch-emacs-event
+  (events &optional buf save-excursion-p)
+  (let ((iiimcf-server-control-enable nil)
+	(iiimcf-server-control-preedit-mode nil))
+    (iiimcf-dispatch-emacs-event events buf save-excursion-p)))
+
 (defun iiimcf-server-control-setup ()
   (if (and iiimcf-server-control-com-id
 	   (not
@@ -231,13 +425,13 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 	     (function iiimp-create-network-channel)
 	     (iiimcf-server-control-parse-hostname
 	      (car iiimcf-server-control-hostlist)))))
+  (iiimcf-server-control-register-com-id
+   iiimcf-server-control-com-id)
   (let ((com-id iiimcf-server-control-com-id)
 	mes)
-    (iiimcf-register-handler
-     (function
-      iiimcf-server-control-setup-event-flow-mode)
-     (list 'iiimp-im-register-trigger-keys
-	   iiimcf-server-control-com-id))
+    (iiimcf-server-control-register-handlers
+     iiimcf-server-control-com-level-handler-alist
+     com-id)
     (if (null iiimcf-server-control-im-id)
 	(progn
 	  (iiimp-send-message
@@ -252,91 +446,18 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 		(nth 2 mes))
 	  (iiimcf-send-client-descriptor
 	   com-id iiimcf-server-control-im-id
-	   (format "IIIMCF-Server-Control/%s"
+	   (format "IIIMCF-SC/%s"
 		   iiimcf-server-control-version))))
     (if (null iiimcf-server-control-ic-id)
-	(progn
-	  (iiimp-send-message
-	   'iiimp-im-createic
-	   com-id iiimcf-server-control-im-id
-	   (if iiimcf-server-control-default-language
-	       (list (cons 'input-language
-			   iiimcf-server-control-default-language))
-	     nil))
-	  (setq mes
-		(iiimcf-message-manager
-		 com-id
-		 (list
-		  (list 'iiimp-im-createic-reply
-			com-id iiimcf-server-control-im-id))))
-	  (setq iiimcf-server-control-ic-id
-		(nth 3 mes))))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-trigger-notify)
-     (list 'iiimp-im-trigger-notify
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-maintain-preedit-text)
-     (list 'iiimp-im-preedit-draw
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-commit-string)
-     (list 'iiimp-im-commit-string
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-forward-event)
-     (list 'iiimp-im-forward-event
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-forward-event)
-     (list 'iiimp-im-forward-event-with-operations
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-prepare-lookup-choice)
-     (list 'iiimp-im-lookup-choice-start
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-draw-lookup-choice)
-     (list 'iiimp-im-lookup-choice-draw
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-redraw-lookup-choice)
-     (list 'iiimp-im-lookup-choice-process
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-clear-lookup-choice)
-     (list 'iiimp-im-lookup-choice-done
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-draw-status)
-     (list 'iiimp-im-status-draw
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
-    (iiimcf-register-handler
-     (function iiimcf-server-control-draw-aux-data)
-     (list 'iiimp-im-aux-draw
-	   com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id))
+	(setq iiimcf-server-control-ic-id
+	      (iiimcf-server-control-ic-setup
+	       com-id iiimcf-server-control-im-id)))
+
+    (iiimcf-server-control-register-handlers
+     iiimcf-server-control-ic-level-handler-alist
+     com-id
+     iiimcf-server-control-im-id
+     iiimcf-server-control-ic-id)
 
     (if iiimcf-server-control-async-invocation-p
 	(iiimp-enable-async-invocation
@@ -344,6 +465,7 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 	 (function iiimcf-server-control-async-handler)))))
 
 (defun iiimcf-server-control-shutdown ()
+  (inactivate-input-method)
   (if (or (null iiimcf-server-control-com-id)
 	  (not (iiimp-check-channel-connection
 		iiimcf-server-control-com-id)))
@@ -351,19 +473,10 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 	    iiimcf-server-control-im-id nil))
   (if iiimcf-server-control-ic-id
       (progn
-	(iiimp-send-message
-	 'iiimp-im-destroyic
+	(iiimcf-server-control-ic-destruct
 	 iiimcf-server-control-com-id
 	 iiimcf-server-control-im-id
 	 iiimcf-server-control-ic-id)
-	(iiimcf-message-manager
-	 iiimcf-server-control-com-id
-	 (list
-	  (list
-	   'iiimp-im-destroyic-reply
-	   iiimcf-server-control-com-id
-	   iiimcf-server-control-im-id
-	   iiimcf-server-control-ic-id)))
 	(setq iiimcf-server-control-ic-id nil)))
   (if iiimcf-server-control-im-id
       (progn
@@ -384,41 +497,15 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 	(while (not (iiimp-destroy-network-channel
 		     iiimcf-server-control-com-id))
 	  (sleep-for 1))
+	(iiimcf-server-control-unregister-com-id
+	 iiimcf-server-control-com-id)
 	(setq iiimcf-server-control-com-id nil)))
   (setq iiimcf-server-control-dynamic-event-flow-p nil)
-  (inactivate-input-method)
 
-  (iiimcf-remove-handler
-   (function
-    iiimcf-server-control-setup-event-flow-mode)
-   nil)
-  (iiimcf-remove-handler
-   (function iiimcf-server-control-maintain-preedit-text)
-   nil)
-  (iiimcf-remove-handler
-   (function iiimcf-server-control-commit-string)
-   nil)
-  (iiimcf-remove-handler
-   (function iiimcf-server-control-forward-event)
-   nil)
-  (iiimcf-remove-handler
-   (function iiimcf-server-control-prepare-lookup-choice)
-   nil)
-  (iiimcf-remove-handler
-   (function iiimcf-server-control-draw-lookup-choice)
-   nil)
-  (iiimcf-remove-handler
-   (function iiimcf-server-control-redraw-lookup-choice)
-   nil)
-  (iiimcf-remove-handler
-   (function iiimcf-server-control-clear-lookup-choice)
-   nil)
-  (iiimcf-remove-handler
-   (function iiimcf-server-control-draw-status)
-   nil)
-  (iiimcf-remove-handler
-   (function iiimcf-server-control-draw-aux-data)
-   nil))
+  (iiimcf-server-control-remove-handlers
+   (append
+    iiimcf-server-control-com-level-handler-alist
+    iiimcf-server-control-ic-level-handler-alist)))
 
 (defun iiimcf-server-control-inactivate ()
   "Inactivate IIIMCF server control input method."
@@ -440,6 +527,28 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 	  iiimcf-server-control-im-id
 	  iiimcf-server-control-ic-id))))
 
+(defun iiimcf-server-control-remove-minor-mode-map ()
+  (let ((islot (assq 'iiimcf-server-control-enable
+		     minor-mode-map-alist))
+	(peslot (assq 'iiimcf-server-control-preedit-mode
+		      minor-mode-map-alist)))
+    (if islot
+	(setq minor-mode-map-alist
+	      (delq islot minor-mode-map-alist)))
+    (if peslot
+	(setq minor-mode-map-alist
+	      (delq peslot minor-mode-map-alist)))))
+
+(defun iiimcf-server-control-register-minor-mode-map ()
+  (iiimcf-server-control-remove-minor-mode-map)
+  (setq minor-mode-map-alist
+	(append
+	 (list (cons 'iiimcf-server-control-preedit-mode
+		     iiimcf-server-control-preedit-state-keymap)
+	       (cons 'iiimcf-server-control-enable
+		     iiimcf-server-control-initial-state-keymap))
+	 minor-mode-map-alist)))
+
 (defun iiimcf-server-control-activate (&optional arg)
   (if (and arg
 	  (< (prefix-numeric-value arg) 0))
@@ -447,18 +556,24 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
       (unwind-protect
 	  (progn
 	    (setq describe-current-input-method-function nil)
-	    (iiimcf-server-control-notify-trigger nil)
+	    (if (iiimp-check-channel-connection
+		 iiimcf-server-control-com-id)
+		(iiimcf-server-control-notify-trigger nil))
 	    (iiimcf-server-control-clear-preedit-text)
 	    ;;(run-hooks 'iiimcf-server-control-inactivate-hook)
 	    )
-	(kill-local-variable 'input-method-function))
+	(setq iiimcf-server-control-enable nil)
+	(iiimcf-server-control-remove-minor-mode-map))
+
     ;; activate
+    (setq iiimcf-server-control-enable t)
     (iiimcf-server-control-setup)
     (iiimcf-server-control-notify-trigger t)
     (setq inactivate-current-input-method-function
 	  'iiimcf-server-control-inactivate)
     (setq describe-current-input-method-function
 	  'iiimcf-server-control-help)
+
     ;; inactivate the current input method also in minibuffers
     ;; before exiting.
     (if (eq (selected-window) (minibuffer-window))
@@ -466,9 +581,7 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 		  (function
 		   iiimcf-server-control-exit-from-minibuffer)))
     ;;(run-hooks 'iiimcf-server-control-activate-hook)
-    (make-local-variable 'input-method-function)
-    (setq input-method-function
-	  (function iiimcf-server-control-input-method-filter))))
+    (iiimcf-server-control-register-minor-mode-map)))
 
 (defun iiimcf-server-control-exit-from-minibuffer ()
   (inactivate-input-method)
@@ -477,34 +590,14 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 		   (function
 		    iiimcf-server-control-exit-from-minibuffer))))
 
-(defun iiimcf-server-control-dispatch-event ()
-  (let ((input-method-function nil)
-	keyseq cmd)
-    (while (input-pending-p)
-      (setq keyseq (read-key-sequence nil nil nil t)
-	    cmd (key-binding keyseq t))
-      (if (and (commandp cmd)
-	       (not (eq cmd 'iiimcf-server-control-keyforward)))
-	  (progn
-	    (setq last-command-event (aref keyseq
-					   (1- (length keyseq)))
-		  last-command this-command
-		  this-command cmd)
-	    (call-interactively cmd))))))
-
-(defun iiimcf-server-control-input-method-filter (&optional ev)
-  (let (iiimcf-server-control-input-method-filtered-events)
-    (iiimcf-server-control-keyforward ev)))
-
 (defun iiimcf-server-control-keyforward (&optional ev)
   (interactive)
   (setq iiimcf-server-control-editing-buffer
 	(current-buffer))
   (let* ((event (or ev
 		    last-command-event))
-	 (keyevent (iiimcf-translate-key event)))
+	 (keyevent (iiimcf-translate-emacs-event event)))
     ;;(if (eq event ? ) (setq event 'convert))
-    (setq keyevent (iiimcf-translate-key event))
     (if keyevent
 	(condition-case err
 	    (progn
@@ -531,12 +624,10 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 	  (iiimp-fatal
 	   (iiimcf-server-control-shutdown)
 	   (signal (car err) (cdr err))))
-      (setq unread-command-events
-	    (nconc
-	     unread-command-events
-	     (list event)))
-      (iiimcf-server-control-dispatch-event))
-    iiimcf-server-control-input-method-filtered-events))
+      ;; Don't send an unattended event to server,
+      ;; Deal with it by myself.
+      (iiimcf-server-control-dispatch-emacs-event (list event)))
+    nil))
 
 (defun iiimcf-server-control-trigger-notify (mes)
   (let ((flag (nth 4 mes)))
@@ -546,7 +637,6 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 
 (defun iiimcf-server-control-forward-event (mes)
   (let ((con (nth 4 mes))
-	(oldbuf (current-buffer))
 	kevslot1 kevslot2 kev)
     (iiimp-add-debug-log
      (format "FWD:%S\n" con))
@@ -561,14 +651,101 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 		      (cons kev kevslot2)))
 	    (setq kevslot1 (cdr kevslot1)))
 	  (if kevslot2
-	      (progn
-		(setq unread-command-events
-		      (nconc (nreverse kevslot2)
-			     unread-command-events))
-		(if iiimcf-server-control-editing-buffer
-		    (set-buffer iiimcf-server-control-editing-buffer))
-		(iiimcf-server-control-dispatch-event)
-		(set-buffer oldbuf)))))))
+	      (iiimcf-server-control-dispatch-emacs-event
+	       (nreverse kevslot2)
+	       (if iiimcf-server-control-editing-buffer
+		   iiimcf-server-control-editing-buffer))
+		)))))
+
+;;;
+;;; Select input method.
+;;;
+
+(defun iiimcf-server-control-completing-read
+  (prompt annotated-collection)
+  (let ((completion-setup-hook))
+    (add-hook 'completion-setup-hook
+	      (lambda ()
+		(let ((inhibit-read-only t)
+		      completion-setup-hook)
+		  (save-excursion
+		    (if (bufferp standard-output)
+			(set-buffer standard-output))
+		    (erase-buffer)
+		    (display-completion-list
+		     (mapcar
+		      (lambda (x)
+			(format "%s (%s)" x (cdr (assoc x annotated-collection))))
+		      (all-completions
+		       (save-excursion
+			 (set-buffer (window-buffer (active-minibuffer-window)))
+			 (buffer-string))
+		       minibuffer-completion-table
+		       minibuffer-completion-predicate
+		       )))))))
+    (completing-read prompt annotated-collection nil t)))
+
+(defun iiimcf-server-control-maintain-language-list (mes)
+  (iiimcf-server-control-com-id-put
+   (nth 1 mes) 'language-list (nth 3 mes)))
+
+(defun iiimcf-server-control-change-language ()
+  (interactive)
+  (inactivate-input-method)
+  (setq iiimcf-server-control-default-language
+	(iiimcf-server-control-completing-read
+	 "Input Language:"
+	 (mapcar
+	  (lambda (x)
+	    (cons x (iiimcf-get-language-description x)))
+	  (iiimcf-server-control-com-id-get
+	   iiimcf-server-control-com-id
+	   'language-list))))
+  (iiimcf-server-control-ic-destruct
+   iiimcf-server-control-com-id
+   iiimcf-server-control-im-id
+   iiimcf-server-control-ic-id)
+  (setq iiimcf-server-control-ic-id
+	(iiimcf-server-control-ic-setup
+	 iiimcf-server-control-com-id
+	 iiimcf-server-control-im-id))
+  (activate-input-method 'iiim-server-control))
+
+(defun iiimcf-server-control-maintain-input-method-list (mes)
+  (iiimcf-server-control-com-id-put
+   (nth 1 mes)
+   'input-method-list
+   (nth 1 (assq 'input-method-list (nth 3 mes)))))
+
+(defun iiimcf-server-control-change-input-method ()
+  (interactive)
+  (inactivate-input-method)
+  (setq iiimcf-server-control-default-input-method
+	(completing-read
+	 "Input Method:"
+	 (mapcar
+	  (lambda (x)
+	    (cons (nth 2 x) 0))
+	  (iiimcf-server-control-com-id-get
+	   iiimcf-server-control-com-id
+	   'input-method-list))
+	  nil t))
+  (iiimcf-server-control-ic-destruct
+   iiimcf-server-control-com-id
+   iiimcf-server-control-im-id
+   iiimcf-server-control-ic-id)
+  (setq iiimcf-server-control-ic-id
+	(iiimcf-server-control-ic-setup
+	 iiimcf-server-control-com-id
+	 iiimcf-server-control-im-id))
+  (activate-input-method 'iiim-server-control))
+
+;;;
+;;; Preedit operations
+;;;
+
+(defsubst iiimcf-server-control-preedit-enable-p ()
+  (> (length iiimcf-server-control-preedit) 0))
 
 (defun iiimcf-server-control-put-face-to-preedit (str)
   (let ((len (length str))
@@ -638,7 +815,7 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 	 '(iiimcf-server-control-preedit-start t intangible t))
 	(prop-end
 	 '(iiimcf-server-control-preedit-end t intangible t
-	   rear-nonsticky t))
+	   rear-nonsticky t start-open t))
 	(oldbuf (current-buffer))
 	pss pse)
     ;; (if (string= " *IIIMP" (buffer-name oldbuf))
@@ -675,7 +852,7 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 	   pts 'iiimcf-server-control-preedit-end))
     (if (null pte)
 	(error "PREEDIT TEXT IS BROKEN!!"))
-    (setq pts (1+ pts))
+    (setq pts (+ pts (length iiimcf-server-control-preedit-open-string)))
     (let ((buffer-undo-list t))
       (delete-region pts pte)
       (goto-char pts)
@@ -684,6 +861,28 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
       (goto-char (+ pts caret)))
     (set-buffer oldbuf)))
 
+(defun iiimcf-server-control-show-preedit-minibuffer-p ()
+  (and input-method-use-echo-area
+       (null
+	(marker-position
+	 iiimcf-server-control-preedit-marker))))
+
+(defun iiimcf-server-control-show-preedit-in-minibuffer (disptext caret)
+  (let (pt mbuf)
+    (save-excursion
+      ;; clear echo area.
+      (message nil)
+      (setq mbuf (window-buffer (minibuffer-window)))
+      (set-buffer mbuf)
+      (erase-buffer)
+      (insert "Preedit:")
+      (setq pt (point))
+      (insert disptext " ")
+      ; (move-overlay
+      ;  iiimcf-server-control-caret-overlay
+      ;  (+ pt caret) (+ 1 pt caret) mbuf)
+      )))
+
 (defun iiimcf-server-control-maintain-preedit-text-internal
   (mes &optional init)
   (let* ((caret (nth 4 mes))
@@ -693,7 +892,7 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 	 (contents (nth 7 mes))
 	 (str (nth 1 contents))
 	 (len (length iiimcf-server-control-preedit))
-	 disptext head tail mbuf pt)
+	 disptext head tail)
     (if (and (>= len ch-first)
 	     (> ch-first 0))
 	(setq head (substring
@@ -708,115 +907,67 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 		    iiimcf-server-control-preedit caret)
 	  caret (cdr disptext)
 	  disptext (car disptext))
-    (if (= 0 (length iiimcf-server-control-preedit))
-	(iiimcf-server-control-clear-preedit-text)
-      (if (and input-method-use-echo-area
-	       (null
-		(marker-position
-		 iiimcf-server-control-preedit-marker)))
-	  ;; show preedit string in minibuffer.
-	  (save-excursion
-	    ;; clear echo area.
-	    (message nil)
-	    (setq mbuf (window-buffer (minibuffer-window)))
-	    (set-buffer mbuf)
-	    (erase-buffer)
-	    (insert "Preedit:")
-	    (setq pt (point))
-	    (insert disptext " ")
-	    (move-overlay
-	     iiimcf-server-control-caret-overlay
-	     (+ pt caret) (+ 1 pt caret) mbuf))
-	(if init
-	    (iiimcf-server-control-setup-preedit-text))
-	(iiimcf-server-control-insert-preedit disptext caret))
-      )))
+
+    (if (iiimcf-server-control-preedit-enable-p)
+	;; Show preedit text.
+	(if (iiimcf-server-control-show-preedit-minibuffer-p)
+	    (iiimcf-server-control-show-preedit-in-minibuffer
+	     disptext caret)
+	  (if init
+	      (iiimcf-server-control-setup-preedit-text))
+	  (iiimcf-server-control-insert-preedit disptext caret))
+      (iiimcf-server-control-clear-preedit-text))))
 
 (defun iiimcf-server-control-maintain-preedit-text (mes)
-  (if iiimcf-server-control-maintaining-preedit-text
-      (iiimcf-server-control-maintain-preedit-text-internal mes)
-    (iiimcf-server-control-maintain-preedit-text-internal mes t)
-    (let ((iiimcf-server-control-maintaining-preedit-text t)
-	  ;; disable the font-lock function invoked by
-	  ;; after-change-functions during preedit.
-	  (font-lock-fontify-region-function (function ignore))
-	  ;; Even read-only text-property cannot protect inserted preedit
-	  ;; text from modifications of the face text-property by font-lock.
-	  keyseq cmd)
-      (while (> (length iiimcf-server-control-preedit) 0)
-	(if iiimcf-server-control-input-method-filtered-events
-	    (save-excursion
-	      (set-buffer
-	       (marker-buffer
-		iiimcf-server-control-preedit-marker))
-	      (goto-char
-	       (marker-position
-		iiimcf-server-control-preedit-marker))
-	      (setq unread-command-events
-		    (nconc unread-command-events
-			   iiimcf-server-control-input-method-filtered-events))
-	      (iiimcf-server-control-dispatch-event)
-	      (setq iiimcf-server-control-input-method-filtered-events nil)))
-	(let ((overriding-terminal-local-map
-	       iiimcf-server-control-keymap)
-	      (input-method-function nil))
-	  (setq keyseq (read-key-sequence
-			nil nil nil t)
-		cmd (key-binding keyseq t)))
-	;; (message "KS:%S, %S" keyseq cmd)
-	(if (commandp cmd)
-	    (progn
-	      (setq last-command-event (aref keyseq
-					     (1- (length keyseq)))
-		    last-command this-command
-		    this-command cmd)
-	      (condition-case err
-		  (if (or (eq cmd
-			      (function
-			       iiimcf-server-control-keyforward))
-			  (null (marker-position
-				 iiimcf-server-control-preedit-marker)))
-		      (call-interactively cmd)
-		    (save-excursion
-		      (set-buffer
-		       (marker-buffer
-			iiimcf-server-control-preedit-marker))
-		      (goto-char
-		       (marker-position
-			iiimcf-server-control-preedit-marker))
-		      (call-interactively cmd)))
-		(error
-		 (debug)
-		 (iiimcf-prevent-error
-		  (message "%s" (cdr err)) (beep))))))))))
+  (let (;; disable the font-lock function invoked by
+	;; after-change-functions during preedit.
+	;; Even read-only text-property cannot protect inserted preedit
+	;; text from modifications of the face text-property by font-lock.
+	(font-lock-fontify-region-function (function ignore)))
+    (if (iiimcf-server-control-preedit-enable-p)
+	(iiimcf-server-control-maintain-preedit-text-internal mes)
+      ;; initialize preedit text.
+      (iiimcf-server-control-maintain-preedit-text-internal mes t)))
+  ;; set iiimcf-server-control-preedit-mode flag
+  (save-excursion
+    (set-buffer iiimcf-server-control-editing-buffer)
+    (setq iiimcf-server-control-preedit-mode
+	  (iiimcf-server-control-preedit-enable-p))))
 
 (defun iiimcf-server-control-clear-preedit-text ()
   (setq iiimcf-server-control-preedit "")
-  (if (marker-position
-       iiimcf-server-control-preedit-marker)
+  (if (iiimcf-server-control-show-preedit-minibuffer-p)
+      ;; erase minibuffer.
+      (save-excursion
+	(let ((mbuf (window-buffer (minibuffer-window))))
+	  (set-buffer mbuf)
+	  (erase-buffer)
+	  ;; (move-overlay
+	  ;; iiimcf-server-control-caret-overlay
+	  ;; 0 0 mbuf)
+	  ))
+    ;; Erase inserted preedit.
       (let ((pts (marker-position
 		  iiimcf-server-control-preedit-marker))
 	    (inhibit-read-only t)
 	    pte)
-	(save-excursion
-	  (set-buffer (marker-buffer
-		       iiimcf-server-control-preedit-marker))
-	  (setq pte
-		(next-single-property-change
-		 pts 'iiimcf-server-control-preedit-end))
-	  (if (null pte)
-	      (error "PREEDIT TEXT BROKEN!!"))
-	  (let ((buffer-undo-list t))
-	    (delete-region pts (1+ pte))))
-	(goto-char pts)
-	(set-marker iiimcf-server-control-preedit-marker nil))
-    (save-excursion
-      (let ((mbuf (window-buffer (minibuffer-window))))
-	(set-buffer mbuf)
-	(erase-buffer)
-	(move-overlay
-	 iiimcf-server-control-caret-overlay
-	 0 0 mbuf)))))
+	(if pts
+	    (progn
+	      (save-excursion
+		(set-buffer (marker-buffer
+			     iiimcf-server-control-preedit-marker))
+		(setq pte
+		      (next-single-property-change
+		       pts 'iiimcf-server-control-preedit-end))
+		(if (null pte)
+		    (error "PREEDIT TEXT BROKEN!!"))
+		(let ((buffer-undo-list t))
+		  (delete-region
+		   pts (+ pte
+			  (length
+			   iiimcf-server-control-preedit-close-string)))))
+	      (goto-char pts)
+	      (set-marker iiimcf-server-control-preedit-marker nil))))))
 
 (defun iiimcf-server-control-prepare-lookup-choice (mes)
   (setq iiimcf-server-control-lookup-choice-configuration mes))
@@ -943,26 +1094,9 @@ If each hostname has the form of \"HOSTNAME:N\", use N-th port."
 	  (set-buffer buf)
 	  (goto-char pt))
       (set-buffer iiimcf-server-control-editing-buffer))
-    (setq generated-events
-	  (append (string-to-list
-		   (if enable-multibyte-characters
-		       str
-		     (string-make-unibyte str)))
-		  generated-events))
-    (if (eq input-method-function
-	    (function iiimcf-server-control-input-method-filter))
-	(setq iiimcf-server-control-input-method-filtered-events
-	      (nconc
-	       (if (and input-method-exit-on-first-char generated-events)
-		   (list (car generated-events))
-		 generated-events)
-	       iiimcf-server-control-input-method-filtered-events))
-      (setq unread-command-events
-	    (nconc
-	     (if (and input-method-exit-on-first-char generated-events)
-		 (list (car generated-events))
-	       generated-events)
-	     unread-command-events)))
+    (setq generated-events (iiimcf-emacs-string-to-events str))
+    ;; dispatch them immediately!
+    (iiimcf-server-control-dispatch-emacs-event generated-events)
     ;; (insert str)
     (if pt (set-marker
 	    iiimcf-server-control-preedit-marker
