@@ -39,6 +39,18 @@
 
 (defvar iiimp-debug-buffer nil)
 
+;; We cannot use such a fundamental function without
+;; dirty aliasing...;_;
+;; Emacs-XEmacs imcompatibility absorption layer...
+(defalias 'iiimp-coding-system-or-name-p
+  (cond ((coding-system-p 'no-conversion)
+	 (function coding-system-p))
+	((find-coding-system 'no-conversion)
+	 (function find-coding-system))
+	(t
+	 (error
+	  "This system have no function to detect coding-system(or its name)."))))
+
 ;; (<COM-ID> <COM-ID> ...)
 (defvar iiimp-communication-list nil
   "IIIMP effective communication list.")
@@ -50,12 +62,17 @@
 (defvar iiimp-process-name "IIIMP")
 
 (defvar iiimp-string-conversion-or-coding-system
-  (if (coding-system-p 'utf-16-be-no-signature-unix)
+  (if (iiimp-coding-system-or-name-p 'utf-16-be-no-signature-unix)
       'utf-16-be-no-signature-unix
     'utf-16-be-unix))
 
+(defvar iiimp-text-warning-char ?!)
+
 (defconst iiimp-basic-opcode-spec-list
-  '((iiimp-im-connect
+  '((iiimp-im-no-protocol
+     0 (nil . nil)
+     "NOP(not written in spec.).")
+    (iiimp-im-connect
      1 (iiimp-send-im-connect . nil)
      "Connection request.")
     (iiimp-im-connect-reply
@@ -290,7 +307,8 @@
 	proc)
     (save-excursion
       (set-buffer buf)
-      (set-buffer-multibyte nil)
+      (if (fboundp (function set-buffer-multibyte))
+	  (set-buffer-multibyte nil))
       (buffer-disable-undo buf)
       (setq proc
 	    (open-network-stream iiimp-process-name buf host port))
@@ -419,16 +437,16 @@
 
 (defconst iiimp-attribid-alist
   '(;; IM attribid
-    (input-method-list . ?\x1001)
-    (object-descriptor-list . ?\x1010)
-    (client-descriptor . ?\x1011)
-    (ccdef . ?\x1030)
-    (jar-gui-object . ?\x1031)
-    (jar-lwe-object . ?\x1032)
-    (binary-gui-object . ?\x1033)
-    (binary-lwe-object . ?\x1034)
-    (script-gui-object . ?\x1035)
-    (script-lwe-object . ?\x1036)
+    (input-method-list . 4097) ; #x1001
+    (object-descriptor-list . 4112) ; #x1010
+    (client-descriptor . 4113) ; #x1011 
+    (ccdef . 4144) ; #x1030
+    (jar-gui-object . 4145) ; #x1031
+    (jar-lwe-object . 4146) ; #x1032
+    (binary-gui-object . 4147) ; #x1033
+    (binary-lwe-object . 4148) ; #x1034
+    (script-gui-object . 4149) ; #x1035
+    (script-lwe-object . 4150) ; #x1036
     ;; IC attribid
     (input-language . 1)
     (character-subsets . 2)
@@ -451,7 +469,7 @@
 (defmacro iiimp-pack-number (bitsspec)
   (let ((sbits 0) (totbits 0) (num 0)
 	prg bls plspec spec val bits ibits
-	mask result)
+	mask tmp)
     (while bitsspec
       (setq plspec nil)
       (while (and (setq spec (car bitsspec))
@@ -459,6 +477,13 @@
 	(setq sbits (+ sbits (cdr spec))
 	      plspec (cons spec plspec)
 	      bitsspec (cdr bitsspec)))
+      ;; Check bitsspec restriction.
+      (if (> sbits 32)
+	  (error "Cannot serialize this %S bitsspec because we need larger integer area." bitsspec)
+	(if (and (> sbits 28)
+		 (> (length plspec) 1))
+	    (error "Cannot serialize more than 28 bits number that is not aligned to byte boundary.%S" bitsspec)))
+      ;; Expand bitsspec chunk.
       (setq ibits sbits
 	    plspec (nreverse plspec)
 	    num 0)
@@ -466,14 +491,21 @@
 	(setq val (car spec)
 	      bits (cdr spec)
 	      ibits (- ibits bits)
-	      mask (1- (lsh 1 bits)))
+	      mask (if (>= bits 28)
+		       nil
+		     (1- (lsh 1 bits))))
 	(if (numberp val)
-	    (setq num (logior (lsh (logand val mask) ibits) num))
+	    (setq num
+		  (if mask
+		      (logior (lsh (logand val mask) ibits) num)
+		    (logior (lsh val ibits) num)))
+	  (setq tmp (if mask
+			(list 'logand val mask)
+		      val))
 	  (setq prg (cons
 		     (if (> ibits 0)
-			 (list 'lsh (list 'logand val mask)
-			       ibits)
-		       (list 'logand val mask))
+			 (list 'lsh tmp ibits)
+		       tmp)
 		     prg)))
 	(setq plspec (cdr plspec)))
       (setq prg
@@ -491,15 +523,18 @@
 	    sbits 0
 	    prg nil))
     (setq bls (cons 'list (nreverse bls)))
+    ;; embedded program
     `(let ((x-ls ,bls)
 	   (x-str (make-string ,(/ (+ totbits 7) 8) 0))
 	   (x-num 0) (x-i 0) (x-sb 0)
 	   x-b x-elem)
        (while (setq x-elem (car x-ls))
-	 (setq x-b (car x-elem)
-	       x-num (logior (lsh x-num x-b) (cdr x-elem))
-	       x-sb (+ x-b x-sb)
-	       x-ls (cdr x-ls))
+	 (setq x-b (car x-elem))
+	 (if (> x-b 28)
+	     (setq x-num (cdr x-elem))
+	   (setq x-num (logior (lsh x-num x-b) (cdr x-elem))))
+	  (setq x-sb (+ x-b x-sb)
+		x-ls (cdr x-ls))
 	 (while (or (>= x-sb 8)
 		    (and (null x-ls)
 			 (> x-sb 0)))
@@ -509,14 +544,12 @@
        x-str)))
 
 (defmacro iiimp-unpack-number (bitsspec packet start)
-  (let ((sbits 0)
-	(num 0)
-	(result
+  (let ((result
 	 (copy-sequence
 	  `(let ((x-i ,start)
 		 (x-num 0)
 		 (x-ib 0)))))
-	spec var bits ib prg)
+	spec var bits)
     (while (setq spec (car bitsspec))
       (setq var (car spec)
 	    bits (cdr spec)
@@ -576,12 +609,13 @@
 (defsubst iiimp-string-pack (str)
   (let* ((bytes (encode-coding-string
 		 str iiimp-string-conversion-or-coding-system))
-	 (len (length bytes)))
+	 len)
     (if (and (not (eq iiimp-string-conversion-or-coding-system
 		      'utf-16-be-no-signature-unix))
-	     (= (aref 0 bytes) ?\xFE)
-	     (= (aref 1 bytes) ?\xFF))
+	     (= (aref bytes 0) ?\xFE)
+	     (= (aref bytes 1) ?\xFF))
 	(setq bytes (substring bytes 2)))
+    (setq len (length bytes))
     (concat (iiimp-pack-number ((len . 16)))
 	    bytes
 	    (iiimp-make-padding (+ len 2)))))
@@ -919,7 +953,7 @@
 ;; How should we deal with surrogate pairs?
 (defun iiimp-text-unpack (packet start)
   (let ((str "") tmpstr
-	len end char
+	len end ucs char
         len-l end-l
 	sidx eidx len-v
 	props props-prev
@@ -933,14 +967,22 @@
     (while (> end start)
       (setq start
 	    (iiimp-unpack-number
-	     ((char . 16)
+	     ((ucs . 16)
 	      (len-l . 16))
 	     packet start))
       (setq end-l (+ start len-l)
 	    props (iiimp-unpack-feedback-list
 		   packet start end-l)
-	    start end-l
-	    tmpstr (char-to-string (ucs-to-char char)))
+	    start end-l)
+      (setq char (ucs-to-char ucs))
+      (if (null char)
+	  (setq tmpstr (char-to-string iiimp-text-warning-char)
+		props (nconc
+		       (list 'unknown-ucs
+			     (cons 'ucs ucs))
+		       props))
+	(setq tmpstr (char-to-string char)))
+
       (setq str (concat str tmpstr))
       (if (equal props props-prev)
 	  (setq ep (length str))
@@ -1202,7 +1244,7 @@
   (let ((type (car im-attribute))
 	(val (nth 1 im-attribute))
 	len)
-    (cond ((eq type 'imput-method-list)
+    (cond ((eq type 'input-method-list)
 	   (setq val
 		 (iiimp-pack-list
 		  (function iiimp-im-descriptor-pack)
@@ -1249,7 +1291,7 @@
 	   (setq val
 		 (iiimp-packet-substring
 		  packet start end)))
-	  ((eq at 'imput-method-list)
+	  ((eq at 'input-method-list)
 	   (setq slot
 		 (iiimp-unpack-list
 		  packet start len
@@ -1457,7 +1499,7 @@
 (defun iiimp-show-message (com-id opcode start len)
   (if iiimp-debug-buffer
       (let* ((end (+ start len))
-	     (str (buffer-substring start end)))
+	     (str (iiimp-packet-substring com-id start end)))
 	(save-excursion
 	  (set-buffer iiimp-debug-buffer)
 	  (goto-char (point-max))
@@ -1465,6 +1507,21 @@
 	  (iiimp-dump-binary str)
 	  (insert "]\n")))))
 
+(defsubst iiimp-parse-header (com-id start)
+  (let (type opcode len)
+    (setq start
+	  (iiimp-unpack-number ((type . 1) (opcode . 7))
+			       com-id start))
+    (setq opcode (aref iiimp-basic-opcode-vector opcode))
+    (setq start
+	  (if (= type 0)
+	      (iiimp-unpack-number ((len . 24))
+				   com-id start)
+	    (iiimp-unpack-number ((len . 56))
+				 com-id start)))
+    (setq len (* 4 len))
+    (cons (list opcode len) start)))
+
 (defsubst iiimp-parse-im-ic-reply-message-internal (com-id opcode start len)
   (let (im-id ic-id)
     (setq start
@@ -1944,19 +2001,10 @@
        ;; Has a header already arrived?
        (if (iiimp-check-channel-bytes com-id start 4 wait)
 	   (progn
-	     ;; parse header
-	     (setq st2
-		   (iiimp-unpack-number ((type . 1) (opcode . 7))
-					com-id start))
-	     (setq opcode (aref iiimp-basic-opcode-vector opcode))
-	     (setq st2
-		   (if (= type 0)
-		       (iiimp-unpack-number ((len . 24))
-					    com-id st2)
-		     (iiimp-unpack-number ((len . 56))
-					  com-id st2)))
-	     (setq len (* 4 len))
-
+	     (setq val (iiimp-parse-header com-id start)
+		   st2 (cdr val)
+		   opcode (car (car val))
+		   len (nth 1 (car val)))
 	     (setq val (funcall
 			(iiimp-select-message-parser opcode)
 			com-id opcode st2 len))
@@ -2119,11 +2167,11 @@
 
 (defun iiimp-send-im-seticfocus (com-id im-id ic-id)
   (iiimp-send-im-ic-request-internal
-   com-id 'iiimp-im-seticforcus im-id ic-id))
+   com-id 'iiimp-im-seticfocus im-id ic-id))
 
 (defun iiimp-send-im-unseticfocus (com-id im-id ic-id)
   (iiimp-send-im-ic-request-internal
-   com-id 'iiimp-im-unseticforcus im-id ic-id))
+   com-id 'iiimp-im-unseticfocus im-id ic-id))
 
 (defun iiimp-send-im-resetic (com-id im-id ic-id)
   (iiimp-send-im-ic-request-internal
@@ -2180,7 +2228,7 @@
     'iiimp-im-preedit-start-reply
     (if (null maxlen)
 	(iiimp-pack-number ((im-id . 16) (ic-id . 16)
-			    (?\xFFFF . 16) (?\xFFFF . 16)))
+			    (65535 . 16) (65535 . 16)))
       (iiimp-pack-number ((im-id . 16) (ic-id . 16)
 			  (maxlen . 32)))))))
 
